"""Wafer environments for agent execution."""

from wafer_core.environments.coding import CodingEnvironment
from wafer_core.environments.gpumode import GPUModeSimpleEnvironment

__all__ = ["CodingEnvironment", "GPUModeSimpleEnvironment"]
