"""ROCprofiler-Systems run module - rocprof-sys-run profiling.

This module provides functionality for the rocprof-sys-run tool,
which is the main comprehensive profiling wrapper.
"""

from wafer_core.lib.rocprofiler.systems.run.profiler import run_systems_profile

__all__ = ["run_systems_profile"]
